package jadex.adapter.jade;

import jade.content.abs.AbsContentElement;
import jade.content.abs.AbsObject;
import jade.content.lang.StringCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.schema.ObjectSchema;
import jadex.runtime.ContentException;
import jadex.runtime.IContentCodec;

import java.io.Serializable;
import java.util.Properties;

/**
 *  A content codec based on JADE codec/ontology implementations.
 */
// Todo: also support ByteArrayCodecs
public class JadeContentCodec implements IContentCodec, Serializable
{
	//-------- attributes --------

	/** The codec (i.e. language). */
	protected StringCodec	codec;

	/** The ontology. */
	protected Ontology	ontology;

	//-------- constructors --------

	/**
	 *  Create a JADE content codec.
	 */
	public JadeContentCodec(StringCodec codec, Ontology ontology)
	{
		this.codec = codec;
		this.ontology = ontology;

		// The following code has been taken from
		// jade.content.ContentManager.getMergedOntology()
		Ontology langonto = codec.getInnerOntology();
		if(langonto != null)
			this.ontology = new Ontology(ontology.getName(), new Ontology[]{this.ontology, langonto}, null);
	}

	//-------- methods --------

	/**
	 *  Test if the codec can be used with the provided meta information.
	 *  @param props The meta information.
	 *  @return True, if it can be used.
	 */
	public boolean match(Properties props)
	{
		String lang = props.getProperty("language"); // Hack!!! avoid dependency to fipa
		String onto = props.getProperty("ontology"); // Hack!!! avoid dependency to fipa
		return lang!=null && lang.equals(codec.getName())
			&& onto!=null && onto.equals(ontology.getName());
	}

	/**
	 *  Encode data with the codec.
	 *  @param val The value.
	 *  @return The encoded object.
	 */
	public synchronized String encode(Object val)
	{
		try
		{
			// The following code has been taken from
			// jade.content.ContentManager.fillContent()
			//AbsContentElement	content	= (AbsContentElement)ontology.fromObject(val);
			//validate(content);
			
			AbsObject	content	= ontology.fromObject(val);
			if(content instanceof AbsContentElement)
			{
				validate((AbsContentElement)content);

				// The following code has been taken from
				// jade.content.ContentManager.encode()
				return codec.encode(ontology, (AbsContentElement)content);
			}
			//else if(content instanceof AbsPrimitive)
			else if(val instanceof String) // Hack?
			{
				return (String)val;
			}
			throw new ContentException("Unable to encode: "+val);
		}
		catch(Exception e)
		{
			throw new ContentException(e.getMessage(), e);
		}
	}

	/**
	 *  Decode data with the codec.
	 *  @param val The string value.
	 *  @return The encoded object.
	 */
	public synchronized Object decode(String val)
	{
		assert val != null;

		try
		{
			// The following code has been taken from
			// jade.content.ContentManager.decode()
			AbsContentElement	content	= codec.decode(ontology, val);

			// The following code has been taken from
			// jade.content.ContentManager.extractContent()
			validate(content);
			return ontology.toObject(content);
		}
		catch(Exception e)
		{
			throw new ContentException(e.getMessage(), e);
		}
	}

	//-------- helper methods --------

	/**
	 *  Validate an abstract object.
	 */
	protected void validate(AbsContentElement content) throws OntologyException
	{
		// The following code has been taken from
		// jade.content.ContentManager.validate()
		ObjectSchema	schema	= ontology.getSchema(content.getTypeName());
		if(schema==null)
		{
			throw new OntologyException("No schema found for type "+content.getTypeName());
		}
		schema.validate(content, ontology);
	}
}
