package jadex.adapter.jade;

import jade.content.ContentElement;
import jade.core.Agent;
import jade.lang.acl.ACLMessage;
import jadex.adapter.IMessageEventTransport;
import jadex.adapter.fipa.SFipa;
import jadex.model.IMMessageEvent;
import jadex.runtime.ContentException;
import jadex.runtime.IContentCodec;
import jadex.runtime.impl.IRMessageEvent;
import jadex.runtime.impl.IRParameter;
import jadex.runtime.impl.IRParameterSet;

import java.io.Serializable;
import java.text.DateFormat;
import java.text.ParseException;
import java.util.Date;
import java.util.Properties;
import java.util.logging.Logger;

/**
 *  Encapsulates the transport layer provided by the JADE platform.
 */
public class JadeMessageEventTransport implements IMessageEventTransport, Serializable
{
	//-------- attributes --------

	/** The JADE agent. */
	protected Agent agent;

	//-------- constructors --------

	/**
	 *  Create a new JADE message event transport.
	 *  @param agent The JADE agent.
	 */
	public JadeMessageEventTransport(Agent agent)
	{
		this.agent = agent;
	}

	//-------- methods --------

	/**
	 *  Send a message event.
	 *  @param msgevent The message event.
	 */
	public boolean sendMessage(IRMessageEvent msgevent)
	{
		IMMessageEvent model = (IMMessageEvent)msgevent.getModelElement();
		if(!model.getMessageType().getName().equals(SFipa.MESSAGE_TYPE_NAME_FIPA))
			return false;

		prepareSending(msgevent);
		ACLMessage aclmsg = (ACLMessage)msgevent.getMessage();
		
		aclmsg.addUserDefinedParameter(JadeMessageAdapter.MESSAGE_ID, msgevent.getId());
		
		if(!aclmsg.getAllReceiver().hasNext())
			throw new RuntimeException("No receiver specified in: "+aclmsg);
		agent.send(aclmsg);
		return true;
	}

	/**
	 *  Prepare the native platform message for sending.
	 */
	public void prepareSending(IRMessageEvent msgevent)
	{
		Logger logger = msgevent.getScope().getLogger();
		Agent agent = (Agent)msgevent.getScope().getAgent().getAgentAdapter();
		
		// Message should be null except in case of forwarding/resending an already received message.
		ACLMessage message = (ACLMessage)msgevent.getMessage();
		if(message==null)
		{
			message	= new ACLMessage(ACLMessage.UNKNOWN);
			msgevent.setMessage(message);
		}		

		// Fill the parameter values in the acl message.
		IRParameter[] params = msgevent.getParameters();
		for(int i=0; i<params.length; i++)
		{
			String pname = params[i].getName();

			if(!SFipa.MESSAGE_ATTRIBUTES.contains(pname))
			{
				logger.warning("Unsupported parameter in message event. Cannot be mapped to acl message: "+msgevent);
			}
			else if(params[i].getValue()!=null)
			{
				if(pname.equals(SFipa.ENCODING))
				{
					if(message.getEncoding()!=null && !message.getEncoding().equals(params[i].getValue()))
						logger.warning(SFipa.ENCODING+" value overriden: "+msgevent);
					message.setEncoding((String)params[i].getValue());
				}
				else if(pname.equals(SFipa.CONVERSATION_ID))
				{
					if(message.getConversationId()!=null && !message.getConversationId().equals(params[i].getValue()))
						logger.warning(SFipa.CONVERSATION_ID+" value overriden: "+msgevent);
					message.setConversationId((String)params[i].getValue());
				}
				else if(pname.equals(SFipa.IN_REPLY_TO))
				{
					if(message.getInReplyTo()!=null && !message.getInReplyTo().equals(params[i].getValue()))
						logger.warning(SFipa.IN_REPLY_TO+" value overriden: "+msgevent);
					message.setInReplyTo((String)params[i].getValue());
				}
				else if(pname.equals(SFipa.LANGUAGE))
				{
					if(message.getLanguage()!=null && !message.getLanguage().equals(params[i].getValue()))
						logger.warning(SFipa.LANGUAGE+" value overriden: "+msgevent);
					message.setLanguage((String)params[i].getValue());
				}
				else if(pname.equals(SFipa.ONTOLOGY))
				{
					if(message.getOntology()!=null && !message.getOntology().equals(params[i].getValue()))
						logger.warning(SFipa.ONTOLOGY+" value overriden: "+msgevent);
					message.setOntology((String)params[i].getValue());
				}
				else if(pname.equals(SFipa.PROTOCOL))
				{
					if(message.getProtocol()!=null && !message.getProtocol().equals(params[i].getValue()))
						logger.warning(SFipa.PROTOCOL+" value overriden: "+msgevent);
					message.setProtocol((String)params[i].getValue());
				}
				else if(pname.equals(SFipa.REPLY_BY))
				{
					Object	val	= params[i].getValue();
					Date	date;
					if(val instanceof String)
					{
						try
						{
							date	= DateFormat.getDateInstance().parse((String)val);
						}
						catch(ParseException e)
						{
							throw new RuntimeException("Error parsing date: "+e);
						}
					}
					else
					{
						date	= (Date)val;
					}

					if(message.getReplyByDate()!=null && !message.getReplyByDate().equals(date))
						logger.warning(SFipa.REPLY_BY+" value overriden: "+msgevent);
					message.setReplyByDate(date);
				}
				else if(pname.equals(SFipa.REPLY_WITH))
				{
					if(message.getReplyWith()!=null && !message.getReplyWith().equals(params[i].getValue()))
						logger.warning(SFipa.REPLY_WITH+" value overriden: "+msgevent);
					message.setReplyWith((String)params[i].getValue());
				}
				else if(pname.equals(SFipa.PERFORMATIVE))
				{
					if(message.getPerformative()!=ACLMessage.UNKNOWN
					&& message.getPerformative()!=SJade.getPerformative(params[i].getValue()))
						logger.warning(SFipa.PERFORMATIVE+" value overriden: "+msgevent);
					message.setPerformative(SJade.getPerformative(params[i].getValue()));
				}
				// not necessary because getContent() can be used
				/*else if(pname.equals(SFipa.CONTENT))
				{
					if(message.getContent()!=null)
						logger.warning(SFipa.CONTENT+" content overriden: "+msgevent);
					message.setContent((String)params[i].getValue());
				}*/
				else if(pname.equals(SFipa.SENDER))
				{
					if(message.getSender()!=null && !message.getSender().equals(SJade.getAID(params[i].getValue())))
						logger.warning(SFipa.SENDER+" value overriden: "+msgevent);
					message.setSender(SJade.getAID(params[i].getValue()));
				}
//				// supported as parameters also for convenience.
//				else if(pname.equals(SFipa.RECEIVER))
//				{
//					// todo: check if receiver is the same and then produce no warning
//					if(message.getAllReceiver().hasNext())
//					{
//						message.clearAllReceiver();
//						logger.warning(SFipa.RECEIVER+" values overriden: "+msgevent);
//					}
//					message.addReceiver(SJade.getAID(params[i].getValue()));
//				}
				else if(pname.equals(SFipa.REPLY_TO) && params[i].getValue()!=null)
				{
					// todo: check if replyto is the same and then produce no warning
					if(message.getAllReplyTo().hasNext())
						logger.warning(SFipa.REPLY_TO+" values overriden: "+msgevent);

					// Remove all reply tos.
					message.clearAllReplyTo();
					message.addReplyTo(SJade.getAID(params[i].getValue()));
				}
			}
		}

		IRParameterSet[] paramsets = msgevent.getParameterSets();
		for(int i=0; i<paramsets.length; i++)
		{
			String psname = paramsets[i].getName();
			if(SFipa.MESSAGE_ATTRIBUTES.contains(psname))
			{
				if(psname.equals(SFipa.RECEIVERS))
				{
					// todo: check if receivers are the same and then produce no warning
					if(message.getAllReceiver().hasNext())
						logger.warning(SFipa.RECEIVERS+" values overriden: "+msgevent);

					// Remove all receivers.
					message.clearAllReceiver();

					// Add receivers from message event.
					Object[] res = paramsets[i].getValues();
					for(int j=0; j<res.length; j++)
						message.addReceiver(SJade.getAID(res[j]));
				}
				else if(psname.equals(SFipa.REPLY_TO))
				{
					// todo: check if replytos are the same and then produce no warning
					if(message.getAllReplyTo().hasNext())
						logger.warning(SFipa.REPLY_TO+" values overriden: "+msgevent);

					// Remove all reply tos.
					message.clearAllReplyTo();

					// Add reply tos from message event.
					Object[] res = paramsets[i].getValues();
					for(int j=0; j<res.length; j++)
						message.addReplyTo(SJade.getAID(res[j]));
				}
			}
		}

		// todo: treat content as normal parameter but as LAST parameter!!

		// If content is not null, a native message should be sent.
		if(message.getContent()==null)
		{
			Object cont = msgevent.getContent();
			if(cont!=null)
			{
				Properties props = new Properties();
				String lang = (String)msgevent.getParameter(SFipa.LANGUAGE).getValue();
				String onto = (String)msgevent.getParameter(SFipa.ONTOLOGY).getValue();
				if(lang!=null)
					props.put(SFipa.LANGUAGE, lang);
				if(onto!=null)
					props.put(SFipa.ONTOLOGY, onto);

				IContentCodec codec = msgevent.getScope().getContentCodec(props);

				// todo: require lang+onto==null?

				if(codec!=null)
				{
					if(message.getContent()!=null)
						logger.warning(SFipa.CONTENT+" value overriden: "+msgevent);
					message.setContent(codec.encode(cont));
				}
				else if(cont instanceof ContentElement)
				{
					try
					{
						agent.getContentManager().fillContent(message, (ContentElement)cont);
					}
					catch(Exception e)
					{
//						e.printStackTrace();
						throw new ContentException(e.getMessage(), e);
					}
				}
				else if(cont instanceof String)
				{
					message.setContent((String)cont);
				}
				else
				{
					throw new ContentException("No content codec found for: "+props);
				}
			}
		}

		// set message id for the tracer
		message.addUserDefinedParameter("message-id", agent.getLocalName()+":"+msgevent.getName());
   	}
}
